/*
What's new:
C44 - [Martin Zoubek] first public release
*/

#ifndef _Types_h
  #define _Types_h

  typedef unsigned long pointer;

  #ifdef UNIX
    #include <stdlib.h>
  #else
    typedef unsigned int uint;
  #endif

  #ifdef WIN32
    #include <windows.h>
  #else
    typedef unsigned char BOOL;
    #define FALSE 0
    #define TRUE 1
  #endif

  typedef double double2[2];
  typedef double double3[3];
  typedef double double4[4];
  typedef double double24[2][4];

  typedef double4 statevec[3];
  typedef double lintransf[3][2];
  typedef double bilintransf[3][2][3];
  typedef double3 viscousvector[3];     /*[3] = DOF */

  typedef double4 matrix4[4];


  #undef min
  #undef max
  #define min(a,b) ((a)<(b)?(a):(b))
  #define max(a,b) ((a)>(b)?(a):(b))
  #define inabs(a) ((a)<(0)?-(a):(a))

  #define pow2(x) ((x)*(x))
  #define pow_fast(a,b) exp((b)*log(a))

  #define double2_dotprod(a,b) ((a)[0]*(b)[0]+(a)[1]*(b)[1])
  #define double2_norm2(a) double2_dotprod(a,a)
  #define double2_sub(c,a,b){\
    (c)[0]=(a)[0]-(b)[0];\
    (c)[1]=(a)[1]-(b)[1];\
    }
  #define double2_crossprod(c,a,b){\
    (c)=(a)[0]*(b)[1]-(a)[1]*(b)[0];\
    } 
  #define double2_distance2(a,b) (pow2((a)[0]-(b)[0])+pow2((a)[1]-(b)[1]))

  #define double2_gravicity(d,a,b,c){\
    (d)[0]=((a)[0]+(b)[0]+(c)[0])/3;\
    (d)[1]=((a)[1]+(b)[1]+(c)[1])/3;\
    }

  #define double2_weight_awer(d,a,b,p,q){\
    (d)[0]=(a)[0]*(p)+(b)[0]*(q);\
    (d)[1]=(a)[1]*(p)+(b)[1]*(q);\
    }

  #define double4_weight_awer(d,a,b,p,q){\
    (d)[0]=(a)[0]*(p)+(b)[0]*(q);\
    (d)[1]=(a)[1]*(p)+(b)[1]*(q);\
    (d)[2]=(a)[2]*(p)+(b)[2]*(q);\
    (d)[3]=(a)[3]*(p)+(b)[3]*(q);\
    }
    #define double4_add(c,a,b){\
    (c)[0]=(a)[0]+(b)[0];\
    (c)[1]=(a)[1]+(b)[1];\
    (c)[2]=(a)[2]+(b)[2];\
    (c)[3]=(a)[3]+(b)[3];\
    }
  #define double4_average(c,a,b){\
    (c)[0]=((a)[0]+(b)[0])/2;\
    (c)[1]=((a)[1]+(b)[1])/2;\
    (c)[2]=((a)[2]+(b)[2])/2;\
    (c)[3]=((a)[3]+(b)[3])/2;\
    }
 
  #define double4_sub(c,a,b){\
    (c)[0]=(a)[0]-(b)[0];\
    (c)[1]=(a)[1]-(b)[1];\
    (c)[2]=(a)[2]-(b)[2];\
    (c)[3]=(a)[3]-(b)[3];\
    }
  #define double4_multiply(c,r,a){\
    (c)[0]=(r)*(a)[0];\
    (c)[1]=(r)*(a)[1];\
    (c)[2]=(r)*(a)[2];\
    (c)[3]=(r)*(a)[3];\
    }
  #define double3_vertex(d,a,b,c){\
    (d)[0]=(a)[0]+(b)[0]-(c)[0];\
    (d)[1]=(a)[1]+(b)[1]-(c)[1];\
    (d)[2]=(a)[2]+(b)[2]-(c)[2];\
    }
  #define double3_vertex4(d,a,b,c){\
    (d)[0]=(a)[1]+(b)[1]-(c)[1];\
    (d)[1]=(a)[2]+(b)[2]-(c)[2];\
    (d)[2]=(a)[3]+(b)[3]-(c)[3];\
    }
  #define double4_aver(d,a,b){\
    (d)[0]=((a)[0]+(b)[0])/2;\
    (d)[1]=((a)[1]+(b)[1])/2;\
    (d)[2]=((a)[2]+(b)[2])/2;\
    (d)[3]=((a)[3]+(b)[3])/2;\
    }

  #define double4_lin_comb(d,a,b,p,q){\
    (d)[0]=(p)*(a)[0]+(q)*(b)[0];\
    (d)[1]=(p)*(a)[1]+(q)*(b)[1];\
    (d)[2]=(p)*(a)[2]+(q)*(b)[2];\
    (d)[3]=(p)*(a)[3]+(q)*(b)[3];\
    }

  #define index2element(p, i, j, k){\
  if((j)>2||(j)<0) printf("index2element-2nd is bad: %d is not in [0,2]\n",j);\
  if((k)>3||(k)<0) printf("index2element-3rd is bad: %d is not in [0,3]\n",k);\
    (p) = 12*(i)+3*(k)+(j);\
  }

  #define element2index( p, i, j, k){\
    (i)= (p)/12;\
    (k)= ((p)-12*(i) )/3;\
    (j)= (p)-12*(i)-3*(k);\
   }

  #define index2matrix(p,row,j,k,stencil) (p) = (row)*(stencil)+(j)*4+(k);

  #define matrix2index( p,row, j, k, stencil){\
    (row) = (p)/(stencil);\
    (j)= ((p)-(stencil)*(row))/4;\
    (k)= (p)-(stencil)*(row)-4*(j);\
   }
/*  #define enum_funct3(F,F_1,F_2){\
    (F)[0]=(4.*(F1)[0]+(F1)[1]+(F1)[2])*(4.*(F2)[0]+(F2)[1]+(F2)[2])/36;\
    (F)[1]=(4.*(F1)[1]+(F1)[2]+(F1)[0])*(4.*(F2)[1]+(F2)[2]+(F2)[0])/36;\
    (F)[2]=(4.*(F1)[2]+(F1)[0]+(F1)[1])*(4.*(F2)[2]+(F2)[0]+(F2)[1])/36;\
    } */

   /* numerical quadratures \int_{K} F1*F2 dx  */
   /* nodes n=3, exact for p=2 */

   /* F1, F2 scalar functions */
  #define volum_integ_scals(I,vol,F1,F2){\
    (I)=(vol)/12*( (F1)[0]*(2.*(F2)[0]+(F2)[1]+(F2)[2]) +                                          (F1)[1]*(2.*(F2)[1]+(F2)[0]+(F2)[2]) +                                          (F1)[2]*(2.*(F2)[2]+(F2)[1]+(F2)[0]) );  \
   }  

   /* F1 vector function, F2 scalar function */
  #define volum_integ_scavec(I,vol,F1,F2){\
    (I)[0]=(vol)/12*( (F1)[0][0]*(2.*(F2)[0]+(F2)[1]+(F2)[2]) +                                       (F1)[1][0]*(2.*(F2)[1]+(F2)[0]+(F2)[2]) +                                       (F1)[2][0]*(2.*(F2)[2]+(F2)[1]+(F2)[0]) );  \
    (I)[1]=(vol)/12*( (F1)[0][1]*(2.*(F2)[0]+(F2)[1]+(F2)[2]) +                                       (F1)[1][1]*(2.*(F2)[1]+(F2)[0]+(F2)[2]) +                                       (F1)[2][1]*(2.*(F2)[2]+(F2)[1]+(F2)[0]) );  \
    (I)[2]=(vol)/12*( (F1)[0][2]*(2.*(F2)[0]+(F2)[1]+(F2)[2]) +                                       (F1)[1][2]*(2.*(F2)[1]+(F2)[0]+(F2)[2]) +                                       (F1)[2][2]*(2.*(F2)[2]+(F2)[1]+(F2)[0]) );  \
    (I)[3]=(vol)/12*( (F1)[0][3]*(2.*(F2)[0]+(F2)[1]+(F2)[2]) +                                       (F1)[1][3]*(2.*(F2)[1]+(F2)[0]+(F2)[2]) +                                       (F1)[2][3]*(2.*(F2)[2]+(F2)[1]+(F2)[0]) );  \
   }  
   /* F1, F2 vector functions */
  #define volum_integ_vects(I,vol,F1,F2){\
    (I)[0]=(vol)/12*( (F1)[0][0]*(2.*(F2)[0][0]+(F2)[1][0]+(F2)[2][0]) +                              (F1)[1][0]*(2.*(F2)[1][0]+(F2)[0][0]+(F2)[2][0]) +                              (F1)[2][0]*(2.*(F2)[2][0]+(F2)[1][0]+(F2)[0][0]) );  \
    (I)[1]=(vol)/12*( (F1)[0][1]*(2.*(F2)[0][1]+(F2)[1][1]+(F2)[2][1]) +                              (F1)[1][1]*(2.*(F2)[1][1]+(F2)[0][1]+(F2)[2][1]) +                              (F1)[2][1]*(2.*(F2)[2][1]+(F2)[1][1]+(F2)[0][1]) );  \
    (I)[2]=(vol)/12*( (F1)[0][2]*(2.*(F2)[0][2]+(F2)[1][2]+(F2)[2][2]) +                              (F1)[1][2]*(2.*(F2)[1][2]+(F2)[0][2]+(F2)[2][2]) +                              (F1)[2][2]*(2.*(F2)[2][2]+(F2)[1][2]+(F2)[0][2]) );  \
    (I)[3]=(vol)/12*( (F1)[0][3]*(2.*(F2)[0][3]+(F2)[1][3]+(F2)[2][3]) +                              (F1)[1][3]*(2.*(F2)[1][3]+(F2)[0][3]+(F2)[2][3]) +                              (F1)[2][3]*(2.*(F2)[2][3]+(F2)[1][3]+(F2)[0][3]) );  \
   }  

  #define barycentre(s,i,x){\
    (x)[0] = ((s)->element[(i)].ve->P[0]->Point[0]+(s)->element[(i)].ve->P[1]->Point[0]+(s)->element[(i)].ve->P[2]->Point[0])/3;\
    (x)[1] = ((s)->element[(i)].ve->P[0]->Point[1]+(s)->element[(i)].ve->P[1]->Point[1]+(s)->element[(i)].ve->P[2]->Point[1])/3;\
  }

  #define Matrix_Times_Vector4(A,x,b){\
    (b)[0]=(A)[0][0]*(x)[0]+(A)[0][1]*(x)[1]+(A)[0][2]*(x)[2]+(A)[0][3]*(x)[3];\
    (b)[1]=(A)[1][0]*(x)[0]+(A)[1][1]*(x)[1]+(A)[1][2]*(x)[2]+(A)[1][3]*(x)[3];\
    (b)[2]=(A)[2][0]*(x)[0]+(A)[2][1]*(x)[1]+(A)[2][2]*(x)[2]+(A)[2][3]*(x)[3];\
    (b)[3]=(A)[3][0]*(x)[0]+(A)[3][1]*(x)[1]+(A)[3][2]*(x)[2]+(A)[3][3]*(x)[3];\
}

  #define vector_w2q(w,q,kappa){\
     (q)[0] = (w)[0];\
     (q)[1] = (w)[1]/(w)[0];\
     (q)[2] = (w)[2]/(w)[0];\
     (q)[3] = ((kappa)-1)*((w)[3]-0.5*((w)[1]*(w)[1]+(w)[2]*(w)[2])/(w)[0]);\
  }
/*     (q)[3] = sqrt(((w)[3]/(w)[0]-0.5*((q)[1]*(q)[1]+(q)[2]*(q)[2]))*kappa*(kappa-1));\  */



  #define vector_q2w(q,w,kappa){\
     (w)[0] = (q)[0];\
     (w)[1] = (q)[1]*(q)[0];\
     (w)[2] = (q)[2]*(q)[0];\
     (w)[3] = (q)[3]/((kappa)-1)+0.5*(q)[0]*((q)[1]*(q)[1]+(q)[2]*(q)[2]);\
  }

/*   (w)[3] = (q)[0]*((q)[3]*(q)[3]/kappa/((kappa)-1)+0.5*((q)[1]*(q)[1]+(q)[2]*(q)[2]));\  */

  #define vector_or_q2w(q,w,kappa){\
    (w)[0] = (q)[0];\
    (w)[1] = (q)[1]*(q)[0];\
    (w)[2] = (q)[2]*(q)[0];\
    (w)[3] = (q)[3]/((kappa)-1)+0.5*(q[0])*((q)[1]*(q)[1]+(q)[2]*(q)[2]);\
  }

#endif

